"""Compute per-portfolio results"""

from itertools import chain
from functools import cached_property

from ..result_computer import ResultComputer
from .portfolio_state import PortfolioState


class PortfolioResultComputer(ResultComputer):
    """Compute per-portfolio results"""

    def __init__(self, portfolios, start_date, end_date, risk_free_return):
        super().__init__(start_date, end_date, risk_free_return)
        self._portfolios = portfolios

    def _perform_todays_computation(self, today, todays_asset_prices):
        """
        Update portfolios state according to any revisions made by the students
        and to compute and add the results based on the stock prices for today.
        """

        # First, incorporate any revisions that the student may have made:
        self._advance_portfolio_revision_states(today)

        # Next, perform the usual steps: processing today's asset prices
        # and adding the results to the output list.
        super()._perform_todays_computation(today, todays_asset_prices)

    def _advance_portfolio_revision_states(self, today):
        """Load new revisions for any portfolio that needs it"""

        for portfolio_state in self._states:
            self._advance_portfolio_revision_state(today, portfolio_state)

    def _advance_portfolio_revision_state(self, today, portfolio_state):
        """
        If this portfolio had an update today, make that the current revision.
        """

        revision_states = portfolio_state.revision_states
        while revision_states and revision_states[0].start_date < today:
            # We need to switch to a newer revision for this portfolio:
            portfolio_state.current_revision_state = revision_states.pop(0)

    def _update_state(self, today, todays_asset_logprices, state):
        """Compute today's returns and update the portfolio state with them."""

        revision_state = state.current_revision_state
        if revision_state:
            todays_returns = self._compute_todays_returns(todays_asset_logprices, revision_state)
            state.process_todays_returns(todays_returns)

    def _add_results(self, today):
        """Add result records based on each state"""

        for state in self._states:
            self._add_result(today, state)

    def _add_result(self, today, state):
        """Add result records for a state"""

        if state.todays_returns is not None:
            result_entry = self._make_result_entry(today, state)
            self._results[state.key].append(result_entry)

    @property
    def _revisions(self):
        """All revisions for all portfolios in one flat iterator"""

        return chain.from_iterable(
            portfolio.portfoliorevision_set.all()
            for portfolio in self._portfolios
        )

    @cached_property
    def _results(self):
        return {portfolio.pk: [] for portfolio in self._portfolios}

    @cached_property
    def _states(self):
        return [
            PortfolioState(portfolio)
            for portfolio in self._portfolios
        ]
