"""Common bits for both result computers"""

from math import log
from functools import cached_property
from itertools import chain, groupby

from ...stocks.models import AssetPrice


class ResultComputer:
    """Common bits for both result computers"""

    def __init__(self, start_date, end_date, risk_free_return):
        self._start_date = start_date
        self._end_date = end_date
        self._risk_free_return = risk_free_return

    @cached_property
    def results(self):
        """Compute and return the results"""

        self._perform_computation()
        return self._results

    def _perform_computation(self):
        """
        Loop through the dates (and prices for that date) to compute and
        add results for that day.
        """

        for today, todays_asset_prices in self._grouped_asset_prices:
            self._perform_todays_computation(today, todays_asset_prices)

    def _perform_todays_computation(self, today, todays_asset_prices):
        """
        Compute and add the results based on the stock prices for today.
        """

        self._process_todays_asset_prices(today, todays_asset_prices)
        self._add_results(today)

    def _process_todays_asset_prices(self, today, asset_prices):
        """
        Maintain the _previous_logprices property and update the
        states with the computed information.
        """
        todays_asset_logprices = self._make_logprices_dict(asset_prices)
        if self._previous_logprices is None:
            # This is the first date for which we have prices.
            # There are no differences to compute, so just initialize
            # _previous_logprices and leave it at that.
            self._previous_logprices = todays_asset_logprices
        else:
            self._update_states(today, todays_asset_logprices)
            self._previous_logprices.update(todays_asset_logprices)

    def _update_states(self, today, todays_asset_logprices):
        """Update each portfolio state according to today's stock prices."""

        for state in self._states:
            self._update_state(today, todays_asset_logprices, state)

    def _compute_todays_returns(self, todays_asset_logprices, revision_state):
        """Compute returns for this portfolio based on the given logprices."""

        asset_weights = revision_state.asset_weights
        asset_returns = list(self._compute_todays_asset_returns(
            todays_asset_logprices,
            asset_weights,
        ))

        if asset_returns:
            return sum(asset_returns)
        else:
            return None

    _previous_logprices = None

    def _compute_todays_asset_returns(self, todays_asset_logprices, asset_weights):
        """
        Compute the returns for this asset based on the difference between
        today's logprices and the previous.
        """

        previous_logprices = self._previous_logprices
        for asset_pk, weight in asset_weights:
            try:
                todays_logprice = todays_asset_logprices[asset_pk]
                previous_logprice = previous_logprices[asset_pk]
            except KeyError:
                pass
            else:
                # We have both values so we can include these in the score:
                yield (todays_logprice - previous_logprice) * weight

    def _add_results(self, today):
        """Add result records based on each state"""

        for state in self._states:
            self._add_result(today, state)

    def _add_result(self, today, state):
        """Add result records for a state"""

        if state.todays_returns is not None:
            result_entry = self._make_result_entry(today, state)
            self._results[state.key].append(result_entry)

    def _make_logprices_dict(self, asset_prices):
        """
        Convert asset prices to logprices and store them in a dict
        (keyed by pk) for easier lookups.
        """

        return {
            asset_price.asset_id: log(asset_price.closing_price)
            for asset_price in asset_prices
        }

    def _make_result_entry(self, today, portfolio_state):
        result_entry = {
            'date': str(today),
            'performance': portfolio_state.todays_returns,
        }

        if portfolio_state.nobs > 1:
            self._extend_result_entry(result_entry, portfolio_state)

        return result_entry

    def _extend_result_entry(self, result_entry, portfolio_state):
        """
        If we have more than one observation we can start adding more data,
        specifically risk and Sharpe ratio.
        """

        mean, standard_deviation = portfolio_state.mean_and_standard_deviation

        result_entry['risk'] = standard_deviation

        if standard_deviation:
            adjusted_mean = mean - self._risk_free_return
            result_entry['sharpe_ratio'] = adjusted_mean / standard_deviation

    @cached_property
    def _grouped_asset_prices(self):
        """Asset prices grouped by date."""

        return groupby(
            self._asset_prices,
            key=lambda asset_price: asset_price.closing_date,
        )

    @property
    def _asset_prices(self):
        """
        Asset prices for all involved assets during the date range between
        start_date and end_date.
        """

        # This list of asset prices can be a bit large, so we should be careful
        # to not turn this iterator into a list or dict.
        return AssetPrice.objects.filter(
            asset__in=self._unique_asset_pks,
            closing_date__gte=self._start_date,
            closing_date__lte=self._end_date,
            closing_price__isnull=False,
        ).order_by('closing_date')

    @property
    def _unique_asset_pks(self):
        """
        Compute the list of all assets referenced by all revisions, using
        a set() to deduplicate.
        """

        return set(
            portfolio_asset.asset_id
            for portfolio_asset in self._assets
        )

    @property
    def _assets(self):
        """All assets for all revisions in one flat iterator"""

        return chain.from_iterable(
            revision.portfolioasset_set.all()
            for revision in self._revisions
        )
