"""Compute results for individual revisions"""

from functools import cached_property

from ..result_computer import ResultComputer
from .revision_state import RevisionState


class RevisionResultComputer(ResultComputer):
    """Compute results for individual revisions"""

    def __init__(self, revisions, start_date, end_date, risk_free_return):
        super().__init__(start_date, end_date, risk_free_return)
        self._revisions = revisions

    # Inactive revisions copy their state from the most recently active one:
    _previous_state = None

    def _update_state(self, today, todays_asset_logprices, state):
        """Compute today's returns and update the state with them."""

        if state.start_date < today:
            todays_returns = self._compute_todays_returns(todays_asset_logprices, state)
            state.process_todays_returns(todays_returns)
            self._previous_state = state
        else:
            self._copy_from_previous_state(state)

    def _copy_from_previous_state(self, state):
        previous_state = self._previous_state
        if previous_state:
            state.copy_from(previous_state)

    @cached_property
    def _results(self):
        return {revision.pk: [] for revision in self._revisions}

    @cached_property
    def _states(self):
        return [
            RevisionState(revision)
            for revision in self._revisions
        ]

