"""Tests for the game app"""

from json import loads as parse_json
from io import BytesIO
from itertools import chain

from openpyxl import load_workbook

from django.contrib.auth import get_user_model
from django.test import TestCase
from django.db import connection
from django.test.utils import CaptureQueriesContext
from rest_framework.test import APIClient

from investment_game.game.models import Game
from investment_game.game.serializers import GameSerializer

User = get_user_model()


class TestGames(TestCase):
    """Game model related tests"""

    fixtures = ['stocks', 'prices', 'players', 'game']
    client_class = APIClient

    def test_exists(self):
        """Dummy test to check if the fixtures work as expected"""

        games = list(Game.objects.all())
        self.assertEqual(len(games), 3)

    def test_serialize(self):
        """Test to check if the serializer works as expected"""

        serializer = GameSerializer(Game(
            name="Foo Bar",
            token='c8ce2f40-3655-43ca-9d72-3910d91686dd',
            start_date='2021-02-10',
            end_date='2021-02-18',
        ))
        self.assertEqual(serializer.data, {
            'pk': None,
            'name': "Foo Bar",
            'token': 'c8ce2f40-3655-43ca-9d72-3910d91686dd',
            'start_date': '2021-02-10',
            'embargo_date': None,
            'end_date': '2021-02-18',
            'risk_free_return': 0.0,
            'index': None,
            'assets': [],
            'max_portfolios': 3,
            'max_portfolio_revisions': 10,
            'max_portfolio_assets': 4,
        })

    def test_admin_ordered_get(self):
        """Test to check if ordering the view works as expected"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        content = parse_json(client.get('/games?ordering=pk').content)
        self.assertEqual(len(content), 3)
        self.assertEqual(content[0]['pk'], 1)
        self.assertEqual(content[1]['pk'], 2)
        self.assertEqual(content[2]['pk'], 3)

    def test_admin_filtered_get(self):
        """Test to check if filtering the view works as expected"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        content = parse_json(client.get('/games?name=Game%20%231').content)
        self.assertEqual(len(content), 1)
        self.assertEqual(content[0]['pk'], 1)

    def test_admin_search_get(self):
        """Test to check if searching the view works as expected"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        content = parse_json(client.get('/games?q=%20%233').content)
        self.assertEqual(len(content), 1)
        self.assertEqual(content[0]['pk'], 3)

    def test_player_get(self):
        """Test to check if the view works as expected for players"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        content = parse_json(client.get('/games').content)
        self.assertEqual(len(content), 1)

    def test_player_join(self):
        """Test to check if a user can join a game."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000920'))

        # Verify user does not yet have access to game
        response = client.get('/games/3')
        self.assertEqual(response.status_code, 404)

        # Join a game
        response = client.post('/join', {
            'token': 'e1bd9ef9-02d3-55f9-a36b-262b12e1c603',
        }, format='json')
        self.assertEqual(response.status_code, 200)

        # Verify user joined the right game
        content = parse_json(response.content)
        self.assertEqual(content['game'], 3)

        # Verify user now has access to game
        response = client.get('/games/3')
        self.assertEqual(response.status_code, 200)

        # Verify joining is idempotent
        response = client.post('/join', {
            'token': 'e1bd9ef9-02d3-55f9-a36b-262b12e1c603',
        }, format='json')
        self.assertEqual(response.status_code, 200)

    def test_player_join_validation_zeroes(self):
        """Test to check if a user can join a game with a nonexistent uuid."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000920'))

        # Join a game
        response = client.post('/join', {
            'token': '00000000-0000-0000-0000-000000000000',
        }, format='json')
        self.assertTrue(400 <= response.status_code < 500)

    def test_player_join_validation_pindakaas(self):
        """Test to check if a user can join a game with a string as uuid."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000920'))

        # Join a game
        response = client.post('/join', {
            'token': 'boterham-met-pindakaas',
        }, format='json')
        self.assertTrue(400 <= response.status_code < 500)

    def test_player_join_validation_empty_string(self):
        """Test to check if a user can join a game with an empty string as uuid."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000920'))

        # Join a game
        response = client.post('/join', {
            'token': '',
        }, format='json')
        self.assertTrue(400 <= response.status_code < 500)

    def test_player_join_validation_null(self):
        """Test to check if a user can join a game with a null uuid."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000920'))

        # Join a game
        response = client.post('/join', {
            'token': None,
        }, format='json')
        self.assertTrue(400 <= response.status_code < 500)

    def test_player_join_validation_number(self):
        """Test to check if a user can join a game with a number as uuid."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000920'))

        # Join a game
        response = client.post('/join', {
            'token': 3,
        }, format='json')
        self.assertTrue(400 <= response.status_code < 500)

    def test_player_join_validation_missing(self):
        """Test to check if a user can join a game with an empty request."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000920'))

        # Join a game
        response = client.post('/join', {
        }, format='json')
        self.assertTrue(400 <= response.status_code < 500)

    def test_player_join_blocked(self):
        """Test to check if a blocked user can rejoin a game."""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        # Join a game
        response = client.post('/join', {
            'token': 'e1bd9ef9-02d3-55f9-a36b-262b12e1c603',
        }, format='json')
        self.assertEqual(response.status_code, 403)

    def test_admin_create_date_validation(self):
        """Creating a game (by an admin) with an incorrect set of dates"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/games', {
            'name': "Foo",
            'start_date': '2021-01-01',
            'end_date': '2020-01-01',
            'index': None,
            'assets': [],
        }, format='json')
        self.assertEqual(
            response.status_code,
            400,
            "Start date before end date",
        )

    def test_admin_create_embargo_date_validation(self):
        """Creating a game (by an admin) with an incorrect set of dates"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/games', {
            'name': "Foo",
            'start_date': '2020-01-01',
            'end_date': '2021-01-01',
            'embargo_date': '2022-01-01',
            'index': None,
            'assets': [],
        }, format='json')
        self.assertEqual(
            response.status_code,
            400,
            "Embargo date before end date",
        )

    def test_player_create_permissions(self):
        """Test permissions"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000900'))

        response = client.post('/games', {
            'name': "Foo",
            'start_date': '2021-02-21',
            'end_date': '2021-03-21',
            'index': None,
            'assets': [],
        }, format='json')
        self.assertEqual(
            response.status_code,
            403,
            "Game creation by players should be rejected",
        )

    def test_player_list_permissions(self):
        """Test permissions"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.get('/games')
        self.assertTrue(
            200 <= response.status_code < 300,
            "Game listing by players should be accepted",
        )

        content = parse_json(response.content)
        self.assertEqual(
            len(content),
            1,
            "Players should only see their own games",
        )

    def test_admin_create_permissions(self):
        """Test permissions"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/games', {
            'name': "Foo",
            'start_date': '2021-02-21',
            'end_date': '2021-03-21',
            'index': None,
            'assets': [],
        }, format='json')
        self.assertTrue(
            200 <= response.status_code < 300,
            "Game creation by admins should be accepted.",
        )

    def test_admin_download_game(self):
        """Test downloads of game information by the admin"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/games/1/download')
        xlsx = response.content
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response['Content-Type'],
            'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
        )
        self.assertIsInstance(xlsx, bytes)

        # with open('download.xlsx', 'wb') as fh:
        #     fh.write(xlsx)

        with BytesIO(xlsx) as fh:
            workbook = load_workbook(fh, read_only=True)
            worksheet = workbook.worksheets[0]
            self.assertEqual(
                worksheet.cell(row=2, column=1).value,
                "Dana Dagschotel",
            )

    def test_admin_download_nonexistent_game(self):
        """Test downloads of game information by the admin"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/games/9001/download')
        self.assertEqual(response.status_code, 404)

    def test_results(self):
        """Fetch results for all participants of this game"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/games/2/results')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)
        self.assertTrue(all(
            portfolio['current']['date'] <= '2021-02-05'
            for portfolio in chain.from_iterable(
                user['portfolios']
                for user in content
            )
        ))
        self.maxDiff = None
        self.assertEqual(content, [
            {
                "pk": 6,
                "portfolios": [
                    {
                        "pk": 11,
                        "name": "Portfolio #1",
                        "current": {
                            "date": "2021-02-05",
                            "performance": -0.07030872574133618,
                            "sharpe_ratio": -0.1895590403699598,
                            "risk": 0.08530567190636686,
                        },
                        "previous": {
                            "date": "2021-02-04",
                            "performance": 0.020730212812393845,
                            "sharpe_ratio": -0.15996265225237533,
                            "risk": 0.08637405998173037,
                        }
                    }
                ]
            },
            {
                "pk": 9,
                "portfolios": [
                    {
                        "pk": 12,
                        "name": "Portfolio #1",
                        "current": {
                            "date": "2021-02-05",
                            "performance": -0.056208293747033844,
                            "sharpe_ratio": 0.001128771294547913,
                            "risk": 0.06594948488034315,
                        },
                        "previous": {
                            "date": "2021-02-04",
                            "performance": -0.04743334426778323,
                            "sharpe_ratio": 0.03803613378174269,
                            "risk": 0.06629268269411277,
                        }
                    },
                    {
                        "pk": 13,
                        "name": "Portfolio #2",
                        "current": {
                            "date": "2021-02-05",
                            "performance": 0.028779093271375178,
                            "sharpe_ratio": -0.13967023341926554,
                            "risk": 0.05835672569594447,
                        },
                        "previous": {
                            "date": "2021-02-04",
                            "performance": -0.040974083566827746,
                            "sharpe_ratio": -0.16510817011365572,
                            "risk": 0.05909059842294174,
                        }
                    }
                ],
                "name": "Dana Dagschotel",
            },
        ])

    def test_results_date(self):
        """Fetch results for a specific date of this game"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/games/2/results?date=2021-02-01')
        self.assertEqual(response.status_code, 200, response.content)
        self.assertTrue(all(
            portfolio['current']['date'] <= '2021-02-01'
            for portfolio in chain.from_iterable(
                user['portfolios']
                for user in parse_json(response.content)
            )
        ))

    def test_results_invalid_date(self):
        """The requested date must not be later than the embargo/end date"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/games/2/results?date=2021-02-10')
        self.assertEqual(response.status_code, 400)

    def test_results_admin(self):
        """Admin may fetch results for any game"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/games/2/results')
        self.assertEqual(response.status_code, 200, response.content)
        self.assertTrue(any(
            portfolio['current']['date'] > '2021-02-05'
            for portfolio in chain.from_iterable(
                user['portfolios']
                for user in parse_json(response.content)
            )
        ))

    def test_results_other(self):
        """Players may not fetch results other games"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.get('/games/2/results')
        self.assertEqual(response.status_code, 404, response.content)

    def test_results_queries(self):
        """Fetching results should not use too many queries"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        with CaptureQueriesContext(connection) as capture:
            response = client.get('/games/3/results')
            self.assertEqual(response.status_code, 200, response.content)

        self.assertLessEqual(len(capture.captured_queries), 10)
        # for q in capture.captured_queries:
        #     print(q['sql'])

    def test_results_blocked(self):
        """Blocked players should not show up in results"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.get('/games/3/results')
        self.assertEqual(response.status_code, 200, response.content)

        content = parse_json(response.content)
        self.assertFalse(any(user['pk'] == 9 for user in content))

    def test_results_blocked_user(self):
        """Blocked players should not be able to see the leaderboard"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/games/3/results')
        self.assertEqual(response.status_code, 404, response.content)
