"""Tests for the game app"""

from json import loads as parse_json
from itertools import chain

from django.contrib.auth import get_user_model
from django.test import TestCase
from rest_framework.test import APIClient

User = get_user_model()


class TestParticipations(TestCase):
    """Participation related tests"""

    fixtures = ['stocks', 'prices', 'players', 'game']
    client_class = APIClient

    def test_admin_list(self):
        """GET-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/games/1/participations')
        self.assertEqual(response.status_code, 200)
        content = parse_json(response.content)
        self.assertIsNotNone(content)
        for participation in content:
            self.assertEqual(participation['game'], 1)

    def test_player_list(self):
        """GET-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/games/1/participations')
        self.assertEqual(response.status_code, 200)
        content = parse_json(response.content)
        self.assertIsNotNone(content)
        self.assertEqual(
            len(content),
            1,
            "Players should only see their own participations",
        )
        for participation in content:
            self.assertEqual(participation['game'], 1)

    def test_non_participant_player_list(self):
        """GET-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.get('/games/2/participations')
        self.assertEqual(response.status_code, 200)
        content = parse_json(response.content)
        self.assertIsNotNone(content)
        self.assertEqual(
            len(content),
            0,
            "Players should not see participations for other games",
        )

    def test_admin_create(self):
        """POST-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/games/1/participations', {
            'game': 1,
            'user': 1,
        }, format='json')
        self.assertTrue(200 <= response.status_code < 300, response.content)

    def test_player_create(self):
        """POST-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000909'))

        response = client.post('/games/1/participations', {
            'game': 1,
            'user': 6,
        }, format='json')
        self.assertEquals(
            response.status_code,
            403,
            "Players cannot create games except by using /join",
        )

    def test_admin_update(self):
        """PUT-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.put('/games/1/participations/1', {
            'game': 2, # should be ignored
            'user': 2, # should be ignored
            'blocked': True,
        }, format='json')
        self.assertTrue(200 <= response.status_code < 300, response.content)
        content = parse_json(response.content)
        self.assertEqual(content['game'], 1)

        response = client.get('/games/1/participations/1')
        self.assertEqual(response.status_code, 200, response.content)

        self.assertEqual(parse_json(response.content), {
            'pk': 1,
            'game': 1,
            'user': 5,
            'blocked': True,
        })

    def test_player_update(self):
        """PUT-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000909'))

        response = client.put('/games/1/participations/1', {
            'game': 2,
            'user': 2,
            'blocked': False,
        }, format='json')
        self.assertEquals(
            response.status_code,
            403,
            "Players cannot update games",
        )

    def test_admin_delete(self):
        """DELETE-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.delete('/games/1/participations/1')
        self.assertTrue(200 <= response.status_code < 300, response.content)

    def test_player_delete(self):
        """DELETE-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.delete('/games/1/participations/3')
        # FIXME: ask PO if this should be allowed
        # self.assertTrue(200 <= response.status_code < 300, response.content)
        self.assertEqual(
            response.status_code,
            403,
            "Players cannot delete games",
        )

    def test_results(self):
        """Fetch results for this participation"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/participations/5/results')
        self.assertEqual(response.status_code, 200, response.content)

        content = parse_json(response.content)
        self.assertTrue(all(
            result['date'] <= '2021-02-05'
            for result in chain.from_iterable(content.values())
        ))
        self.maxDiff = None
        self.assertEqual(content, {
            "12": [
                {
                    "date": "2021-01-05",
                    "performance": 0.08934745518227549
                }, {
                    "date": "2021-01-06",
                    "performance": -0.005495442118975458,
                    "sharpe_ratio": 0.8841148409559821,
                    "risk": 0.047421448650625475
                }, {
                    "date": "2021-01-07",
                    "performance": 0.11283010784280512,
                    "sharpe_ratio": 1.2817167402761922,
                    "risk": 0.0511506988311429
                }, {
                    "date": "2021-01-08",
                    "performance": 0.10333144144992845,
                    "sharpe_ratio": 1.5883605019278952,
                    "risk": 0.04722063442019111
                }, {
                    "date": "2021-01-11",
                    "performance": -0.06693377422606517,
                    "sharpe_ratio": 0.6587739915688827,
                    "risk": 0.07076168492167838
                }, {
                    "date": "2021-01-12",
                    "performance": -0.18357341711103548,
                    "sharpe_ratio": 0.07683470151817144,
                    "risk": 0.10738717888475752
                }, {
                    "date": "2021-01-13",
                    "performance": -0.0009207382812872122,
                    "sharpe_ratio": 0.0697757662937032,
                    "risk": 0.09947299822679119
                }, {
                    "date": "2021-01-14",
                    "performance": -0.019614919430171768,
                    "sharpe_ratio": 0.038746638086525474,
                    "risk": 0.09346202257205906
                }, {
                    "date": "2021-01-15",
                    "performance": -0.00731961527795634,
                    "sharpe_ratio": 0.027280235627347615,
                    "risk": 0.08818389957100202
                }, {
                    "date": "2021-01-18",
                    "performance": 0.08794282374391951,
                    "sharpe_ratio": 0.12524198125458258,
                    "risk": 0.0875057394298664
                }, {
                    "date": "2021-01-19",
                    "performance": -0.08037202054013305,
                    "sharpe_ratio": 0.03037179511260368,
                    "risk": 0.08746721969859825
                }, {
                    "date": "2021-01-20",
                    "performance": 0.004660383427054793,
                    "sharpe_ratio": 0.033715600716183396,
                    "risk": 0.08374531853858654
                }, {
                    "date": "2021-01-21",
                    "performance": 0.017492398333820304,
                    "sharpe_ratio": 0.049058518974398765,
                    "risk": 0.08055478996262078
                }, {
                    "date": "2021-01-22",
                    "performance": 0.008369496752494498,
                    "sharpe_ratio": 0.054969514108658475,
                    "risk": 0.07763287487026396
                }, {
                    "date": "2021-01-25",
                    "performance": -0.026212923202439554,
                    "sharpe_ratio": 0.029653388189129017,
                    "risk": 0.0753848797544306
                }, {
                    "date": "2021-01-26",
                    "performance": 0.011930865689872538,
                    "sharpe_ratio": 0.03890769131253244,
                    "risk": 0.0730288162514654
                }, {
                    "date": "2021-01-27",
                    "performance": 0.006025306920833451,
                    "sharpe_ratio": 0.042746269156311634,
                    "risk": 0.07085231911483857
                }, {
                    "date": "2021-01-28",
                    "performance": 0.017452061838226385,
                    "sharpe_ratio": 0.05555893789717534,
                    "risk": 0.06893529406431119
                }, {
                    "date": "2021-01-29",
                    "performance": 0.0350577483023935,
                    "sharpe_ratio": 0.08113988607768684,
                    "risk": 0.06745805515492134
                }, {
                    "date": "2021-02-01",
                    "performance": 0.06974963007091661,
                    "sharpe_ratio": 0.12922641922517153,
                    "risk": 0.06722575399374417
                }, {
                    "date": "2021-02-02",
                    "performance": -0.08287128201109946,
                    "sharpe_ratio": 0.06322758848086978,
                    "risk": 0.06844178349446199
                }, {
                    "date": "2021-02-03",
                    "performance": 0.01455265590959555,
                    "sharpe_ratio": 0.07162991974603967,
                    "risk": 0.0669021114734047
                }, {
                    "date": "2021-02-04",
                    "performance": -0.04743334426778323,
                    "sharpe_ratio": 0.03803613378174269,
                    "risk": 0.06629268269411277
                }, {
                    "date": "2021-02-05",
                    "performance": -0.056208293747033844,
                    "sharpe_ratio": 0.001128771294547913,
                    "risk": 0.06594948488034315
                }
            ],
            "13": [
                {
                    "date": "2021-01-05",
                    "performance": 0.08664309370379086
                }, {
                    "date": "2021-01-06",
                    "performance": 0.0422715632572265,
                    "sharpe_ratio": 2.905346190757753,
                    "risk": 0.02218576522328217
                }, {
                    "date": "2021-01-07",
                    "performance": -0.02795267887301596,
                    "sharpe_ratio": 0.7133302183314859,
                    "risk": 0.04717870045477506
                }, {
                    "date": "2021-01-08",
                    "performance": -0.04928591676364146,
                    "sharpe_ratio": 0.23748871313763656,
                    "risk": 0.054398439236995534
                }, {
                    "date": "2021-01-11",
                    "performance": -0.05113797245799444,
                    "sharpe_ratio": 0.0019570491873470925,
                    "risk": 0.05498981526314325
                }, {
                    "date": "2021-01-12",
                    "performance": -0.014435181285449739,
                    "sharpe_ratio": -0.045873771815901014,
                    "risk": 0.050490334196686325
                }, {
                    "date": "2021-01-13",
                    "performance": 0.009693182240038789,
                    "sharpe_ratio": -0.01279595137224522,
                    "risk": 0.04693348540766734
                }, {
                    "date": "2021-01-14",
                    "performance": -0.022485323906060506,
                    "sharpe_ratio": -0.07497842165153494,
                    "risk": 0.04449485848265981
                }, {
                    "date": "2021-01-15",
                    "performance": -0.005777314098678321,
                    "sharpe_ratio": -0.08597801745703981,
                    "risk": 0.041957169394434855
                }, {
                    "date": "2021-01-18",
                    "performance": -0.0029583532516135325,
                    "sharpe_ratio": -0.08899713345567256,
                    "risk": 0.039804542078922284
                }, {
                    "date": "2021-01-19",
                    "performance": 0.018771914487044405,
                    "sharpe_ratio": -0.03933201597141414,
                    "risk": 0.038490473142694916
                }, {
                    "date": "2021-01-20",
                    "performance": -0.00683302171846667,
                    "sharpe_ratio": -0.053066897643430866,
                    "risk": 0.03688113448900568
                }, {
                    "date": "2021-01-21",
                    "performance": 0.017432629871341865,
                    "sharpe_ratio": -0.013003575766565188,
                    "risk": 0.03580895986520556
                }, {
                    "date": "2021-01-22",
                    "performance": 0.04213195004217076,
                    "sharpe_ratio": 0.07117259371710288,
                    "risk": 0.036208330605154215
                }, {
                    "date": "2021-01-25",
                    "performance": -0.024069946232249326,
                    "sharpe_ratio": 0.02248396447252776,
                    "risk": 0.035606487545426974
                }, {
                    "date": "2021-01-26",
                    "performance": -0.005142828274719372,
                    "sharpe_ratio": 0.012435933221129321,
                    "risk": 0.034505837929690426
                }, {
                    "date": "2021-01-27",
                    "performance": 0.010957915301024865,
                    "sharpe_ratio": 0.031234539532325235,
                    "risk": 0.033567123612331225
                }, {
                    "date": "2021-01-28",
                    "performance": -0.09721850089585021,
                    "sharpe_ratio": -0.11129013730723743,
                    "risk": 0.03963353545772244
                }, {
                    "date": "2021-01-29",
                    "performance": -0.047601403843153804,
                    "sharpe_ratio": -0.16809306998075274,
                    "risk": 0.03976374601748174
                }, {
                    "date": "2021-02-01",
                    "performance": -0.011843296016729218,
                    "sharpe_ratio": -0.17904047153806518,
                    "risk": 0.03877321354280123
                }, {
                    "date": "2021-02-02",
                    "performance": -0.1784875270668176,
                    "sharpe_ratio": -0.2872979594634888,
                    "risk": 0.052596302123211786
                }, {
                    "date": "2021-02-03",
                    "performance": 0.1339052660883759,
                    "sharpe_ratio": -0.1388767864957642,
                    "risk": 0.060034167474392074
                }, {
                    "date": "2021-02-04",
                    "performance": -0.040974083566827746,
                    "sharpe_ratio": -0.16510817011365572,
                    "risk": 0.05909059842294174
                }, {
                    "date": "2021-02-05",
                    "performance": 0.028779093271375178,
                    "sharpe_ratio": -0.13967023341926554,
                    "risk": 0.05835672569594447
                }
            ]
        })

    def test_results_admin(self):
        """Admin may fetch results for any participation"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/participations/5/results')
        self.assertEqual(response.status_code, 200, response.content)

        self.assertTrue(any(
            result['date'] > '2021-02-05'
            for result in chain.from_iterable(
                parse_json(response.content).values()
            )
        ))

    def test_results_other(self):
        """Players may not fetch results for other people's participations"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.get('/participations/5/results')
        self.assertEqual(response.status_code, 404, response.content)

    def test_results_blocked_user(self):
        """Blocked players should not be able to see the participation results"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/participations/8/results')
        self.assertEqual(response.status_code, 404, response.content)

    def test_results_risk_free_return(self):
        """Games with a risk free rate should have lower results"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/participations/5/results')
        self.assertEqual(response.status_code, 200, response.content)
        no_rfr_response = parse_json(response.content)

        response = client.get('/games/2')
        self.assertEqual(response.status_code, 200, response.content)
        game = parse_json(response.content)

        game['risk_free_return'] = 0.1
        response = client.put('/games/2', game, format='json')
        self.assertEqual(response.status_code, 200, response.content)

        response = client.get('/participations/5/results')
        self.assertEqual(response.status_code, 200, response.content)
        rfr_response = parse_json(response.content)

        self.assertLess(
            rfr_response['12'][1]['sharpe_ratio'],
            no_rfr_response['12'][1]['sharpe_ratio'],
        )
