"""Tests for the game app"""

from json import loads as parse_json
from itertools import chain

from freezegun import freeze_time

from django.contrib.auth import get_user_model
from django.test import TestCase
from rest_framework.test import APIClient


User = get_user_model()


class TestPortfolios(TestCase):
    """Portfolio related tests"""

    fixtures = ['stocks', 'prices', 'players', 'game']
    client_class = APIClient

    def test_admin_retrieve(self):
        """POST-related tests"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        content = parse_json(client.get('/portfolios/1').content)
        self.assertIsInstance(content['last_revision'], dict)

    def test_admin_retrieve_empty(self):
        """
        Test if we can fetch a newly created portfolio (which does not have
        revisions yet, so we should get ‘null’ for last_revision).
        """

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/participations/1/portfolios', {
            'name': "Testportfolio",
        }, format='json')
        self.assertTrue(200 <= response.status_code < 300, response.content)

        content = parse_json(response.content)
        path = f'/portfolios/{content["pk"]}'

        content = parse_json(client.get(path).content)
        self.assertIs(content['last_revision'], None)

    @freeze_time('2021-01-01')
    def test_admin_create(self):
        """An admin may create a portfolio for any game"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/participations/5/portfolios', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 201, response.content)

    @freeze_time('2021-01-01')
    def test_player_create(self):
        """A player may create a portfolio for their participation"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.post('/participations/5/portfolios', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 201, response.content)

    @freeze_time('2021-02-01')
    def test_player_create_after_game_start(self):
        """A player may not create a portfolio when the game has started"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.post('/participations/5/portfolios', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 400)

    @freeze_time('2021-02-01')
    def test_admin_create_after_game_start(self):
        """An admin may not create a portfolio when the game has started"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/participations/5/portfolios', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 400)

    def test_player_create_other(self):
        """A player may not create a portfolio for someone else's game"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.post('/participations/5/portfolios', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 403)

    @freeze_time('2021-02-01')
    def test_admin_update(self):
        """An admin may update any portfolio"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.put('/portfolios/13', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 200, response.content)

    @freeze_time('2021-02-05')
    def test_admin_update_after_embargo(self):
        """An admin may not update a portfolio after the embargo starts"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.put('/portfolios/13', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 400)

    @freeze_time('2021-02-20')
    def test_admin_update_after_game_ends(self):
        """An admin may not update a portfolio after the game finishes"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.put('/portfolios/13', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 400)

    @freeze_time('2021-02-01')
    def test_player_update(self):
        """A player may update their portfolio"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.put('/portfolios/13', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 200, response.content)

    @freeze_time('2021-02-05')
    def test_player_update_after_embargo(self):
        """A player may not update their portfolio after the embargo starts"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.put('/portfolios/13', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 400)

    @freeze_time('2021-02-20')
    def test_player_update_after_game_end(self):
        """A player may not update their portfolio after the game finishes"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.put('/portfolios/13', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 400)

    def test_player_update_other(self):
        """A player may not update someone else's portfolio"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.put('/portfolios/13', {
            'name': "New name",
        }, format='json')
        self.assertEqual(response.status_code, 404)

    def test_results(self):
        """Fetch results for this portfolio"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/portfolios/13/results')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)

        self.assertTrue(all(
            result['date'] <= '2021-02-05'
            for result in chain.from_iterable(content.values())
        ))
        self.assertEqual(content, {
            "46": [
                {
                    "date": "2021-01-05",
                    "performance": 0.08664309370379086
                }, {
                    "date": "2021-01-06",
                    "performance": 0.0422715632572265,
                    "sharpe_ratio": 2.905346190757753,
                    "risk": 0.02218576522328217
                }, {
                    "date": "2021-01-07",
                    "performance": -0.02795267887301596,
                    "sharpe_ratio": 0.7133302183314859,
                    "risk": 0.04717870045477506
                }, {
                    "date": "2021-01-08",
                    "performance": 0.06121124384257533,
                    "sharpe_ratio": 0.9525083537567078,
                    "risk": 0.042564776805096516
                }, {
                    "date": "2021-01-11",
                    "performance": -0.11424664319461462,
                    "sharpe_ratio": 0.13187614822408925,
                    "risk": 0.07268422589128602
                }, {
                    "date": "2021-01-12",
                    "performance": -0.00917807693650412,
                    "sharpe_ratio": 0.09679559435192393,
                    "risk": 0.06671877657740356
                }, {
                    "date": "2021-01-13",
                    "performance": -0.055073980770972586,
                    "sharpe_ratio": -0.035652649758137055,
                    "risk": 0.06541480920678716
                }, {
                    "date": "2021-01-14",
                    "performance": -0.03352624257304065,
                    "sharpe_ratio": -0.10042081689900655,
                    "risk": 0.062053520231133
                }, {
                    "date": "2021-01-15",
                    "performance": -0.11028196785598847,
                    "sharpe_ratio": -0.26547075866733316,
                    "risk": 0.06702294537046273
                }, {
                    "date": "2021-01-18",
                    "performance": 0.07324656752030888,
                    "sharpe_ratio": -0.12555732823431934,
                    "risk": 0.06920115544198516
                }, {
                    "date": "2021-01-19",
                    "performance": -0.015524647965021728,
                    "sharpe_ratio": -0.14104173294319752,
                    "risk": 0.06600997237301993
                }, {
                    "date": "2021-01-20",
                    "performance": 0.07655805652874419,
                    "sharpe_ratio": -0.03191396633302114,
                    "risk": 0.06750887956788199
                }, {
                    "date": "2021-01-21",
                    "performance": 0.016448278505208004,
                    "sharpe_ratio": -0.011122207801649974,
                    "risk": 0.06504958353481054
                }, {
                    "date": "2021-01-22",
                    "performance": 0.09119174682919921,
                    "sharpe_ratio": 0.08718683255741816,
                    "risk": 0.06700414797156873
                }, {
                    "date": "2021-01-25",
                    "performance": 0.00937563892259114,
                    "sharpe_ratio": 0.09387760743339954,
                    "risk": 0.06473815814217453
                }, {
                    "date": "2021-01-26",
                    "performance": -0.01771813021930167,
                    "sharpe_ratio": 0.07292280926670348,
                    "risk": 0.0629465436292499
                }, {
                    "date": "2021-01-27",
                    "performance": 0.011831088772892899,
                    "sharpe_ratio": 0.08210998479356182,
                    "risk": 0.0610908790121191
                }, {
                    "date": "2021-01-28",
                    "performance": 0.028866753888293375,
                    "sharpe_ratio": 0.106359450021469,
                    "risk": 0.0596204993534925
                }, {
                    "date": "2021-01-29",
                    "performance": 0.11375244278365365,
                    "sharpe_ratio": 0.1910197199010458,
                    "risk": 0.06279156228755102
                }, {
                    "date": "2021-02-01",
                    "performance": -0.04436888673929929,
                    "sharpe_ratio": 0.14700298920049526,
                    "risk": 0.06242227468463841
                }, {
                    "date": "2021-02-02",
                    "performance": -0.04846848228987858,
                    "sharpe_ratio": 0.10349235158071969,
                    "risk": 0.06214249747699048
                }, {
                    "date": "2021-02-03",
                    "performance": 0.06318976241202108,
                    "sharpe_ratio": 0.14568472543008806,
                    "risk": 0.061854147703857026
                }, {
                    "date": "2021-02-04",
                    "performance": -0.008996549699969833,
                    "sharpe_ratio": 0.1357665996940725,
                    "risk": 0.06060591270329758
                }, {
                    "date": "2021-02-05",
                    "performance": 0.025012035658862234,
                    "sharpe_ratio": 0.1502338523197034,
                    "risk": 0.0594245743661805
                }
            ],
            "47": [
                {
                    "date": "2021-01-05",
                    "performance": 0.08664309370379086
                }, {
                    "date": "2021-01-06",
                    "performance": 0.0422715632572265,
                    "sharpe_ratio": 2.905346190757753,
                    "risk": 0.02218576522328217
                }, {
                    "date": "2021-01-07",
                    "performance": -0.02795267887301596,
                    "sharpe_ratio": 0.7133302183314859,
                    "risk": 0.04717870045477506
                }, {
                    "date": "2021-01-08",
                    "performance": -0.04928591676364146,
                    "sharpe_ratio": 0.23748871313763656,
                    "risk": 0.054398439236995534
                }, {
                    "date": "2021-01-11",
                    "performance": -0.05113797245799444,
                    "sharpe_ratio": 0.0019570491873470925,
                    "risk": 0.05498981526314325
                }, {
                    "date": "2021-01-12",
                    "performance": -0.12286879906038456,
                    "sharpe_ratio": -0.29994876937886505,
                    "risk": 0.06797311334606791
                }, {
                    "date": "2021-01-13",
                    "performance": 0.08590214161322862,
                    "sharpe_ratio": -0.07119075794434024,
                    "risk": 0.07310051720331398
                }, {
                    "date": "2021-01-14",
                    "performance": 0.049483959270421055,
                    "sharpe_ratio": 0.023072332755076266,
                    "risk": 0.07073076890522823
                }, {
                    "date": "2021-01-15",
                    "performance": 0.0008292019679827824,
                    "sharpe_ratio": 0.023134249122563934,
                    "risk": 0.06668608561009656
                }, {
                    "date": "2021-01-18",
                    "performance": 0.040251483083505055,
                    "sharpe_ratio": 0.08416555079674012,
                    "risk": 0.06432094274753469
                }, {
                    "date": "2021-01-19",
                    "performance": 0.1585370604918836,
                    "sharpe_ratio": 0.2561103892621925,
                    "risk": 0.07549057861895012
                }, {
                    "date": "2021-01-20",
                    "performance": -0.07289784715717346,
                    "sharpe_ratio": 0.1519818415514301,
                    "risk": 0.076640344908424
                }, {
                    "date": "2021-01-21",
                    "performance": -0.03983740086845724,
                    "sharpe_ratio": 0.10263611007813976,
                    "risk": 0.07490083028529389
                }, {
                    "date": "2021-01-22",
                    "performance": -0.011707339084507412,
                    "sharpe_ratio": 0.08710822306345828,
                    "risk": 0.07234887658784443
                }, {
                    "date": "2021-01-25",
                    "performance": -0.003010252874571309,
                    "sharpe_ratio": 0.08123849894004355,
                    "risk": 0.06993424493739744
                }, {
                    "date": "2021-01-26",
                    "performance": 0.027303568945512623,
                    "sharpe_ratio": 0.10355132190482047,
                    "risk": 0.06791551711022091
                }, {
                    "date": "2021-01-27",
                    "performance": 0.026687829982196633,
                    "sharpe_ratio": 0.1239809753218628,
                    "risk": 0.06604983727856782
                }, {
                    "date": "2021-01-28",
                    "performance": -0.0987791828589375,
                    "sharpe_ratio": 0.032693446368556596,
                    "risk": 0.06870645140800204
                }, {
                    "date": "2021-01-29",
                    "performance": -0.06815930250659619,
                    "sharpe_ratio": -0.021242609437428138,
                    "risk": 0.06869705678654811
                }, {
                    "date": "2021-02-01",
                    "performance": -0.07172841096493482,
                    "sharpe_ratio": -0.07239770188936044,
                    "risk": 0.06868671142797873
                }, {
                    "date": "2021-02-02",
                    "performance": 0.010860264737894126,
                    "sharpe_ratio": -0.0628583228773734,
                    "risk": 0.06711611609901656
                }, {
                    "date": "2021-02-03",
                    "performance": -0.08032379110923253,
                    "sharpe_ratio": -0.11381402417280197,
                    "risk": 0.0674620200300458
                }, {
                    "date": "2021-02-04",
                    "performance": -0.0033470748011691608,
                    "sharpe_ratio": -0.11350776231786207,
                    "risk": 0.06598506868193726
                }, {
                    "date": "2021-02-05",
                    "performance": -0.0044141197968977225,
                    "sharpe_ratio": -0.11395997101883508,
                    "risk": 0.06459867755328996
                }
            ],
            "48": [
                {
                    "date": "2021-01-05",
                    "performance": 0.08664309370379086
                }, {
                    "date": "2021-01-06",
                    "performance": 0.0422715632572265,
                    "sharpe_ratio": 2.905346190757753,
                    "risk": 0.02218576522328217
                }, {
                    "date": "2021-01-07",
                    "performance": -0.02795267887301596,
                    "sharpe_ratio": 0.7133302183314859,
                    "risk": 0.04717870045477506
                }, {
                    "date": "2021-01-08",
                    "performance": -0.04928591676364146,
                    "sharpe_ratio": 0.23748871313763656,
                    "risk": 0.054398439236995534
                }, {
                    "date": "2021-01-11",
                    "performance": -0.05113797245799444,
                    "sharpe_ratio": 0.0019570491873470925,
                    "risk": 0.05498981526314325
                }, {
                    "date": "2021-01-12",
                    "performance": -0.014435181285449739,
                    "sharpe_ratio": -0.045873771815901014,
                    "risk": 0.050490334196686325
                }, {
                    "date": "2021-01-13",
                    "performance": 0.009693182240038789,
                    "sharpe_ratio": -0.01279595137224522,
                    "risk": 0.04693348540766734
                }, {
                    "date": "2021-01-14",
                    "performance": -0.022485323906060506,
                    "sharpe_ratio": -0.07497842165153494,
                    "risk": 0.04449485848265981
                }, {
                    "date": "2021-01-15",
                    "performance": -0.005777314098678321,
                    "sharpe_ratio": -0.08597801745703981,
                    "risk": 0.041957169394434855
                }, {
                    "date": "2021-01-18",
                    "performance": -0.0029583532516135325,
                    "sharpe_ratio": -0.08899713345567256,
                    "risk": 0.039804542078922284
                }, {
                    "date": "2021-01-19",
                    "performance": 0.018771914487044405,
                    "sharpe_ratio": -0.03933201597141414,
                    "risk": 0.038490473142694916
                }, {
                    "date": "2021-01-20",
                    "performance": -0.00683302171846667,
                    "sharpe_ratio": -0.053066897643430866,
                    "risk": 0.03688113448900568
                }, {
                    "date": "2021-01-21",
                    "performance": -0.0012258902920199155,
                    "sharpe_ratio": -0.05364546099788085,
                    "risk": 0.03543478365487891
                }, {
                    "date": "2021-01-22",
                    "performance": 0.005635513001890358,
                    "sharpe_ratio": -0.03984099157228008,
                    "risk": 0.03420093082906652
                }, {
                    "date": "2021-01-25",
                    "performance": 0.015801658920899353,
                    "sharpe_ratio": -0.00655256944409866,
                    "risk": 0.033317485239824904
                }, {
                    "date": "2021-01-26",
                    "performance": -0.022578400763667572,
                    "sharpe_ratio": -0.04939772657789443,
                    "risk": 0.032710422106863674
                }, {
                    "date": "2021-01-27",
                    "performance": 0.013625233042330764,
                    "sharpe_ratio": -0.02252296515648013,
                    "risk": 0.03193575632283809
                }, {
                    "date": "2021-01-28",
                    "performance": -0.004062749739693672,
                    "sharpe_ratio": -0.029151986926831432,
                    "risk": 0.03104542437072726
                }, {
                    "date": "2021-01-29",
                    "performance": -0.011022177649229858,
                    "sharpe_ratio": -0.04744009771246123,
                    "risk": 0.03030172837716105
                }, {
                    "date": "2021-02-01",
                    "performance": 0.028584105947287775,
                    "sharpe_ratio": 0.002101256326445589,
                    "risk": 0.030250564506987668
                }, {
                    "date": "2021-02-02",
                    "performance": 0.018822699890351124,
                    "sharpe_ratio": 0.03211940218957922,
                    "risk": 0.0297906032187663
                }, {
                    "date": "2021-02-03",
                    "performance": 0.0004817048354670561,
                    "sharpe_ratio": 0.032133022719598135,
                    "risk": 0.029105838487748874
                }, {
                    "date": "2021-02-04",
                    "performance": 0.02530004598723079,
                    "sharpe_ratio": 0.06902564957835343,
                    "risk": 0.028896463343233127
                }, {
                    "date": "2021-02-05",
                    "performance": 0.004172513256186949,
                    "sharpe_ratio": 0.07370946433415601,
                    "risk": 0.028291396171846693
                }
            ],
            "49": [
                {
                    "date": "2021-01-05",
                    "performance": 0.08664309370379086
                }, {
                    "date": "2021-01-06",
                    "performance": 0.0422715632572265,
                    "sharpe_ratio": 2.905346190757753,
                    "risk": 0.02218576522328217
                }, {
                    "date": "2021-01-07",
                    "performance": -0.02795267887301596,
                    "sharpe_ratio": 0.7133302183314859,
                    "risk": 0.04717870045477506
                }, {
                    "date": "2021-01-08",
                    "performance": -0.04928591676364146,
                    "sharpe_ratio": 0.23748871313763656,
                    "risk": 0.054398439236995534
                }, {
                    "date": "2021-01-11",
                    "performance": -0.05113797245799444,
                    "sharpe_ratio": 0.0019570491873470925,
                    "risk": 0.05498981526314325
                }, {
                    "date": "2021-01-12",
                    "performance": -0.014435181285449739,
                    "sharpe_ratio": -0.045873771815901014,
                    "risk": 0.050490334196686325
                }, {
                    "date": "2021-01-13",
                    "performance": 0.009693182240038789,
                    "sharpe_ratio": -0.01279595137224522,
                    "risk": 0.04693348540766734
                }, {
                    "date": "2021-01-14",
                    "performance": -0.022485323906060506,
                    "sharpe_ratio": -0.07497842165153494,
                    "risk": 0.04449485848265981
                }, {
                    "date": "2021-01-15",
                    "performance": -0.005777314098678321,
                    "sharpe_ratio": -0.08597801745703981,
                    "risk": 0.041957169394434855
                }, {
                    "date": "2021-01-18",
                    "performance": -0.0029583532516135325,
                    "sharpe_ratio": -0.08899713345567256,
                    "risk": 0.039804542078922284
                }, {
                    "date": "2021-01-19",
                    "performance": 0.018771914487044405,
                    "sharpe_ratio": -0.03933201597141414,
                    "risk": 0.038490473142694916
                }, {
                    "date": "2021-01-20",
                    "performance": -0.00683302171846667,
                    "sharpe_ratio": -0.053066897643430866,
                    "risk": 0.03688113448900568
                }, {
                    "date": "2021-01-21",
                    "performance": 0.017432629871341865,
                    "sharpe_ratio": -0.013003575766565188,
                    "risk": 0.03580895986520556
                }, {
                    "date": "2021-01-22",
                    "performance": 0.04213195004217076,
                    "sharpe_ratio": 0.07117259371710288,
                    "risk": 0.036208330605154215
                }, {
                    "date": "2021-01-25",
                    "performance": -0.024069946232249326,
                    "sharpe_ratio": 0.02248396447252776,
                    "risk": 0.035606487545426974
                }, {
                    "date": "2021-01-26",
                    "performance": -0.005142828274719372,
                    "sharpe_ratio": 0.012435933221129321,
                    "risk": 0.034505837929690426
                }, {
                    "date": "2021-01-27",
                    "performance": 0.010957915301024865,
                    "sharpe_ratio": 0.031234539532325235,
                    "risk": 0.033567123612331225
                }, {
                    "date": "2021-01-28",
                    "performance": -0.09721850089585021,
                    "sharpe_ratio": -0.11129013730723743,
                    "risk": 0.03963353545772244
                }, {
                    "date": "2021-01-29",
                    "performance": -0.047601403843153804,
                    "sharpe_ratio": -0.16809306998075274,
                    "risk": 0.03976374601748174
                }, {
                    "date": "2021-02-01",
                    "performance": -0.061900593304281666,
                    "sharpe_ratio": -0.2327332531670527,
                    "risk": 0.04058225101742325
                }, {
                    "date": "2021-02-02",
                    "performance": 0.02969882475728627,
                    "sharpe_ratio": -0.18731109291270304,
                    "risk": 0.04047200397750054
                }, {
                    "date": "2021-02-03",
                    "performance": -0.06920095767749737,
                    "sharpe_ratio": -0.24972654734742067,
                    "risk": 0.04157254866259746
                }, {
                    "date": "2021-02-04",
                    "performance": 0.01713476728721699,
                    "sharpe_ratio": -0.223793056968285,
                    "risk": 0.04104415937487604
                }, {
                    "date": "2021-02-05",
                    "performance": -0.054925991942550115,
                    "sharpe_ratio": -0.26916297253157245,
                    "risk": 0.04120647011029372
                }
            ],
            "50": [
                {
                    "date": "2021-01-05",
                    "performance": 0.08664309370379086
                }, {
                    "date": "2021-01-06",
                    "performance": 0.0422715632572265,
                    "sharpe_ratio": 2.905346190757753,
                    "risk": 0.02218576522328217
                }, {
                    "date": "2021-01-07",
                    "performance": -0.02795267887301596,
                    "sharpe_ratio": 0.7133302183314859,
                    "risk": 0.04717870045477506
                }, {
                    "date": "2021-01-08",
                    "performance": -0.04928591676364146,
                    "sharpe_ratio": 0.23748871313763656,
                    "risk": 0.054398439236995534
                }, {
                    "date": "2021-01-11",
                    "performance": -0.05113797245799444,
                    "sharpe_ratio": 0.0019570491873470925,
                    "risk": 0.05498981526314325
                }, {
                    "date": "2021-01-12",
                    "performance": -0.014435181285449739,
                    "sharpe_ratio": -0.045873771815901014,
                    "risk": 0.050490334196686325
                }, {
                    "date": "2021-01-13",
                    "performance": 0.009693182240038789,
                    "sharpe_ratio": -0.01279595137224522,
                    "risk": 0.04693348540766734
                }, {
                    "date": "2021-01-14",
                    "performance": -0.022485323906060506,
                    "sharpe_ratio": -0.07497842165153494,
                    "risk": 0.04449485848265981
                }, {
                    "date": "2021-01-15",
                    "performance": -0.005777314098678321,
                    "sharpe_ratio": -0.08597801745703981,
                    "risk": 0.041957169394434855
                }, {
                    "date": "2021-01-18",
                    "performance": -0.0029583532516135325,
                    "sharpe_ratio": -0.08899713345567256,
                    "risk": 0.039804542078922284
                }, {
                    "date": "2021-01-19",
                    "performance": 0.018771914487044405,
                    "sharpe_ratio": -0.03933201597141414,
                    "risk": 0.038490473142694916
                }, {
                    "date": "2021-01-20",
                    "performance": -0.00683302171846667,
                    "sharpe_ratio": -0.053066897643430866,
                    "risk": 0.03688113448900568
                }, {
                    "date": "2021-01-21",
                    "performance": 0.017432629871341865,
                    "sharpe_ratio": -0.013003575766565188,
                    "risk": 0.03580895986520556
                }, {
                    "date": "2021-01-22",
                    "performance": 0.04213195004217076,
                    "sharpe_ratio": 0.07117259371710288,
                    "risk": 0.036208330605154215
                }, {
                    "date": "2021-01-25",
                    "performance": -0.024069946232249326,
                    "sharpe_ratio": 0.02248396447252776,
                    "risk": 0.035606487545426974
                }, {
                    "date": "2021-01-26",
                    "performance": -0.005142828274719372,
                    "sharpe_ratio": 0.012435933221129321,
                    "risk": 0.034505837929690426
                }, {
                    "date": "2021-01-27",
                    "performance": 0.010957915301024865,
                    "sharpe_ratio": 0.031234539532325235,
                    "risk": 0.033567123612331225
                }, {
                    "date": "2021-01-28",
                    "performance": -0.09721850089585021,
                    "sharpe_ratio": -0.11129013730723743,
                    "risk": 0.03963353545772244
                }, {
                    "date": "2021-01-29",
                    "performance": -0.047601403843153804,
                    "sharpe_ratio": -0.16809306998075274,
                    "risk": 0.03976374601748174
                }, {
                    "date": "2021-02-01",
                    "performance": -0.011843296016729218,
                    "sharpe_ratio": -0.17904047153806518,
                    "risk": 0.03877321354280123
                }, {
                    "date": "2021-02-02",
                    "performance": 0.1135020605143224,
                    "sharpe_ratio": -0.026393907766343233,
                    "risk": 0.04571298084060446
                }, {
                    "date": "2021-02-03",
                    "performance": -0.1562440033240377,
                    "sharpe_ratio": -0.14975567961653546,
                    "risk": 0.05511444677140351
                }, {
                    "date": "2021-02-04",
                    "performance": 0.055146830878377455,
                    "sharpe_ratio": -0.09916947015295152,
                    "risk": 0.05543194434095193
                }, {
                    "date": "2021-02-05",
                    "performance": -0.06175726882962382,
                    "sharpe_ratio": -0.14149643613428098,
                    "risk": 0.05541714059419577
                }
            ],
            "51": [
                {
                    "date": "2021-01-05",
                    "performance": 0.08664309370379086
                }, {
                    "date": "2021-01-06",
                    "performance": 0.0422715632572265,
                    "sharpe_ratio": 2.905346190757753,
                    "risk": 0.02218576522328217
                }, {
                    "date": "2021-01-07",
                    "performance": -0.02795267887301596,
                    "sharpe_ratio": 0.7133302183314859,
                    "risk": 0.04717870045477506
                }, {
                    "date": "2021-01-08",
                    "performance": -0.04928591676364146,
                    "sharpe_ratio": 0.23748871313763656,
                    "risk": 0.054398439236995534
                }, {
                    "date": "2021-01-11",
                    "performance": -0.05113797245799444,
                    "sharpe_ratio": 0.0019570491873470925,
                    "risk": 0.05498981526314325
                }, {
                    "date": "2021-01-12",
                    "performance": -0.014435181285449739,
                    "sharpe_ratio": -0.045873771815901014,
                    "risk": 0.050490334196686325
                }, {
                    "date": "2021-01-13",
                    "performance": 0.009693182240038789,
                    "sharpe_ratio": -0.01279595137224522,
                    "risk": 0.04693348540766734
                }, {
                    "date": "2021-01-14",
                    "performance": -0.022485323906060506,
                    "sharpe_ratio": -0.07497842165153494,
                    "risk": 0.04449485848265981
                }, {
                    "date": "2021-01-15",
                    "performance": -0.005777314098678321,
                    "sharpe_ratio": -0.08597801745703981,
                    "risk": 0.041957169394434855
                }, {
                    "date": "2021-01-18",
                    "performance": -0.0029583532516135325,
                    "sharpe_ratio": -0.08899713345567256,
                    "risk": 0.039804542078922284
                }, {
                    "date": "2021-01-19",
                    "performance": 0.018771914487044405,
                    "sharpe_ratio": -0.03933201597141414,
                    "risk": 0.038490473142694916
                }, {
                    "date": "2021-01-20",
                    "performance": -0.00683302171846667,
                    "sharpe_ratio": -0.053066897643430866,
                    "risk": 0.03688113448900568
                }, {
                    "date": "2021-01-21",
                    "performance": 0.017432629871341865,
                    "sharpe_ratio": -0.013003575766565188,
                    "risk": 0.03580895986520556
                }, {
                    "date": "2021-01-22",
                    "performance": 0.04213195004217076,
                    "sharpe_ratio": 0.07117259371710288,
                    "risk": 0.036208330605154215
                }, {
                    "date": "2021-01-25",
                    "performance": -0.024069946232249326,
                    "sharpe_ratio": 0.02248396447252776,
                    "risk": 0.035606487545426974
                }, {
                    "date": "2021-01-26",
                    "performance": -0.005142828274719372,
                    "sharpe_ratio": 0.012435933221129321,
                    "risk": 0.034505837929690426
                }, {
                    "date": "2021-01-27",
                    "performance": 0.010957915301024865,
                    "sharpe_ratio": 0.031234539532325235,
                    "risk": 0.033567123612331225
                }, {
                    "date": "2021-01-28",
                    "performance": -0.09721850089585021,
                    "sharpe_ratio": -0.11129013730723743,
                    "risk": 0.03963353545772244
                }, {
                    "date": "2021-01-29",
                    "performance": -0.047601403843153804,
                    "sharpe_ratio": -0.16809306998075274,
                    "risk": 0.03976374601748174
                }, {
                    "date": "2021-02-01",
                    "performance": -0.011843296016729218,
                    "sharpe_ratio": -0.17904047153806518,
                    "risk": 0.03877321354280123
                }, {
                    "date": "2021-02-02",
                    "performance": -0.1784875270668176,
                    "sharpe_ratio": -0.2872979594634888,
                    "risk": 0.052596302123211786
                }, {
                    "date": "2021-02-03",
                    "performance": 0.1339052660883759,
                    "sharpe_ratio": -0.1388767864957642,
                    "risk": 0.060034167474392074
                }, {
                    "date": "2021-02-04",
                    "performance": -0.08468036576314243,
                    "sharpe_ratio": -0.19189860050233093,
                    "risk": 0.06074360982826134
                }, {
                    "date": "2021-02-05",
                    "performance": 0.024216373372353907,
                    "sharpe_ratio": -0.16966154599793903,
                    "risk": 0.05989516281019303
                }
            ],
            "52": [
                {
                    "date": "2021-01-05",
                    "performance": 0.08664309370379086
                }, {
                    "date": "2021-01-06",
                    "performance": 0.0422715632572265,
                    "sharpe_ratio": 2.905346190757753,
                    "risk": 0.02218576522328217
                }, {
                    "date": "2021-01-07",
                    "performance": -0.02795267887301596,
                    "sharpe_ratio": 0.7133302183314859,
                    "risk": 0.04717870045477506
                }, {
                    "date": "2021-01-08",
                    "performance": -0.04928591676364146,
                    "sharpe_ratio": 0.23748871313763656,
                    "risk": 0.054398439236995534
                }, {
                    "date": "2021-01-11",
                    "performance": -0.05113797245799444,
                    "sharpe_ratio": 0.0019570491873470925,
                    "risk": 0.05498981526314325
                }, {
                    "date": "2021-01-12",
                    "performance": -0.014435181285449739,
                    "sharpe_ratio": -0.045873771815901014,
                    "risk": 0.050490334196686325
                }, {
                    "date": "2021-01-13",
                    "performance": 0.009693182240038789,
                    "sharpe_ratio": -0.01279595137224522,
                    "risk": 0.04693348540766734
                }, {
                    "date": "2021-01-14",
                    "performance": -0.022485323906060506,
                    "sharpe_ratio": -0.07497842165153494,
                    "risk": 0.04449485848265981
                }, {
                    "date": "2021-01-15",
                    "performance": -0.005777314098678321,
                    "sharpe_ratio": -0.08597801745703981,
                    "risk": 0.041957169394434855
                }, {
                    "date": "2021-01-18",
                    "performance": -0.0029583532516135325,
                    "sharpe_ratio": -0.08899713345567256,
                    "risk": 0.039804542078922284
                }, {
                    "date": "2021-01-19",
                    "performance": 0.018771914487044405,
                    "sharpe_ratio": -0.03933201597141414,
                    "risk": 0.038490473142694916
                }, {
                    "date": "2021-01-20",
                    "performance": -0.00683302171846667,
                    "sharpe_ratio": -0.053066897643430866,
                    "risk": 0.03688113448900568
                }, {
                    "date": "2021-01-21",
                    "performance": 0.017432629871341865,
                    "sharpe_ratio": -0.013003575766565188,
                    "risk": 0.03580895986520556
                }, {
                    "date": "2021-01-22",
                    "performance": 0.04213195004217076,
                    "sharpe_ratio": 0.07117259371710288,
                    "risk": 0.036208330605154215
                }, {
                    "date": "2021-01-25",
                    "performance": -0.024069946232249326,
                    "sharpe_ratio": 0.02248396447252776,
                    "risk": 0.035606487545426974
                }, {
                    "date": "2021-01-26",
                    "performance": -0.005142828274719372,
                    "sharpe_ratio": 0.012435933221129321,
                    "risk": 0.034505837929690426
                }, {
                    "date": "2021-01-27",
                    "performance": 0.010957915301024865,
                    "sharpe_ratio": 0.031234539532325235,
                    "risk": 0.033567123612331225
                }, {
                    "date": "2021-01-28",
                    "performance": -0.09721850089585021,
                    "sharpe_ratio": -0.11129013730723743,
                    "risk": 0.03963353545772244
                }, {
                    "date": "2021-01-29",
                    "performance": -0.047601403843153804,
                    "sharpe_ratio": -0.16809306998075274,
                    "risk": 0.03976374601748174
                }, {
                    "date": "2021-02-01",
                    "performance": -0.011843296016729218,
                    "sharpe_ratio": -0.17904047153806518,
                    "risk": 0.03877321354280123
                }, {
                    "date": "2021-02-02",
                    "performance": -0.1784875270668176,
                    "sharpe_ratio": -0.2872979594634888,
                    "risk": 0.052596302123211786
                }, {
                    "date": "2021-02-03",
                    "performance": 0.1339052660883759,
                    "sharpe_ratio": -0.1388767864957642,
                    "risk": 0.060034167474392074
                }, {
                    "date": "2021-02-04",
                    "performance": -0.040974083566827746,
                    "sharpe_ratio": -0.16510817011365572,
                    "risk": 0.05909059842294174
                }, {
                    "date": "2021-02-05",
                    "performance": 0.028779093271375178,
                    "sharpe_ratio": -0.13967023341926554,
                    "risk": 0.05835672569594447
                }
            ]
        })

    def test_results_admin(self):
        """Admin may fetch results for any portfolio"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/portfolios/13/results')
        self.assertEqual(response.status_code, 200, response.content)

        self.assertTrue(any(
            result['date'] > '2021-02-05'
            for result in chain.from_iterable(
                parse_json(response.content).values()
            )
        ))

    def test_results_other(self):
        """Player may not see results of someone else's portfolio"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.get('/portfolios/13/results')
        self.assertEqual(response.status_code, 404, response.content)

    def test_results_blocked_user(self):
        """Blocked players should not be able to see the portfolio results"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/portfolios/18/results')
        self.assertEqual(response.status_code, 404, response.content)
