"""Tests for the game app"""

from json import loads as parse_json
from datetime import date

from freezegun import freeze_time

from django.contrib.auth import get_user_model
from django.test import TestCase
from rest_framework.test import APIClient

User = get_user_model()


class TestPortfolioRevisions(TestCase):
    """Portfolio revision related tests"""

    fixtures = ['stocks', 'prices', 'players', 'game']
    client_class = APIClient

    def test_player_retrieve(self):
        """Player may retrieve their portfolio revisions"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/portfolio-revisions/38')
        self.assertTrue(200 <= response.status_code < 300, response.content)

    def test_admin_retrieve(self):
        """Admin may retrieve any portfolio revision"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/portfolio-revisions/16')
        self.assertTrue(200 <= response.status_code < 300, response.content)

    def test_player_retrieve_other(self):
        """Players may not retrieve other people's portfolio revisions"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000920'))

        response = client.get('/portfolio-revisions/16')
        self.assertEqual(response.status_code, 404, response.content)

    def test_admin_list(self):
        """Admin may retrieve any portfolio revision"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        content = parse_json(
            client.get('/portfolios/12/revisions?ordering=serial').content
        )
        self.assertGreater(
            len(content),
            1,
            "Test result is sufficiently large",
        )

        dates = [revision['date'] for revision in content]

        # Test if the dates are unique
        self.assertEqual(len(set(dates)), len(content))

        # Test if the dates are sorted
        self.assertEqual(sorted(dates), dates)

    def test_admin_create(self):
        """POST-related test"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/portfolios/5/revisions', {
            'assets': [
                {'asset': 1, 'weight': 60},
                {'asset': 2, 'weight': 40}
            ]
        }, format='json')
        self.assertTrue(200 <= response.status_code < 300, response.content)

    def test_admin_update(self):
        """PUT-related test"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.put('/portfolios/5/revisions/23', {
            'assets': [
                {'asset': 1, 'weight': 60},
                {'asset': 2, 'weight': 40}
            ]
        }, format='json')
        self.assertTrue(200 <= response.status_code < 300, response.content)

    @freeze_time('2021-02-04')
    def test_player_create(self):
        """Player should be able to create a revision"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.post('/portfolios/13/revisions', {
            'assets': [
                {'asset': 1, 'weight': 40},
                {'asset': 3, 'weight': 60}
            ]
        }, format='json')
        self.assertTrue(200 <= response.status_code < 300, response.content)

    @freeze_time('2021-02-04')
    def test_player_create_same_day(self):
        """Two revisions cannot be created on the same day"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        content = parse_json(client.get('/portfolios/13').content)
        initial_revision_count = content['revision_count']

        response = client.post('/portfolios/13/revisions', {
            'assets': [
                {'asset': 1, 'weight': 40},
                {'asset': 3, 'weight': 60}
            ]
        }, format='json')

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], initial_revision_count + 1)

        response = client.post('/portfolios/13/revisions', {
            'assets': [
                {'asset': 3, 'weight': 40},
                {'asset': 1, 'weight': 60}
            ]
        }, format='json')

        self.assertEqual(response.status_code, 400)

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], initial_revision_count + 1)

    @freeze_time('2021-02-04')
    def test_player_create_same_assets(self):
        """
        Trying to create a revision with the same assets/weights as the
        last should just return that last revision instead
        """

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        content = parse_json(client.get('/portfolios/13').content)
        initial_revision_count = content['revision_count']

        # If this assertion fails, pick a different game for this test:
        self.assertLess(content['last_revision']['date'], str(date.today()))

        response = client.post('/portfolios/13/revisions', {
            'assets': content['last_revision']['assets'],
        }, format='json')

        self.assertEqual(response.status_code, 400)

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], initial_revision_count)

    @freeze_time('2021-02-04')
    def test_player_create_other(self):
        """
        Player should not be able to create a revision in someone else's
        portfolio.
        """

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000901'))

        response = client.post('/portfolios/1/revisions', {
            'assets': [
                {'asset': 1, 'weight': 40},
                {'asset': 2, 'weight': 60}
            ]
        }, format='json')

        self.assertEqual(
            response.status_code,
            403,
            "Players should not be able to create revisions on other "
            "people's portfolios",
        )

    @freeze_time('2021-01-01')
    def test_player_create_before_game_start(self):
        """
        The initial revision may be created/updated without restriction
        if the game hasn't started yet.
        """

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.post('/portfolios/13/revisions', {
            'assets': [
                {'asset': 1, 'weight': 40},
                {'asset': 3, 'weight': 60}
            ]
        }, format='json')
        self.assertTrue(200 <= response.status_code < 300, response.content)

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], 1)

    @freeze_time('2021-02-05')
    def test_player_create_after_embargo(self):
        """Player should not be able to create a revision after the embargo starts"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.post('/portfolios/12/revisions', {
            'assets': [
                {'asset': 1, 'weight': 40},
                {'asset': 3, 'weight': 60}
            ]
        }, format='json')
        self.assertEqual(response.status_code, 400)

    @freeze_time('2021-02-20')
    def test_player_create_after_game_end(self):
        """Player should not be able to create a revision after the game finishes"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.post('/portfolios/12/revisions', {
            'assets': [
                {'asset': 1, 'weight': 40},
                {'asset': 3, 'weight': 60}
            ]
        }, format='json')
        self.assertEqual(response.status_code, 400)

    @freeze_time('2021-02-04')
    def test_player_update(self):
        """Players may not update their revisions after the game has started"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.put('/portfolios/12/revisions/38', {
            'assets': [
                {'asset': 1, 'weight': 60},
                {'asset': 2, 'weight': 40}
            ]
        }, format='json')
        self.assertEqual(response.status_code, 400)

    @freeze_time('2021-02-04')
    def test_player_update_today(self):
        """Revisions may not be updated even if they were created today"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        content = parse_json(client.get('/portfolios/13').content)
        initial_revision_count = content['revision_count']

        response = client.post('/portfolios/13/revisions', {
            'assets': [
                {'asset': 1, 'weight': 40},
                {'asset': 3, 'weight': 60}
            ]
        }, format='json')

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], initial_revision_count + 1)
        pk = content['last_revision']['pk']

        response = client.put(f'/portfolios/13/revisions/{pk}', {
            'assets': [
                {'asset': 3, 'weight': 10},
                {'asset': 1, 'weight': 90}
            ]
        }, format='json')

        self.assertEqual(response.status_code, 400)

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], initial_revision_count + 1)

    @freeze_time('2021-01-01')
    def test_player_update_today_before_game_start(self):
        """
        Revisions may be updated even if they were created today,
        as long as the game hasn't started yet.
        """

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        content = parse_json(client.get('/portfolios/13').content)
        initial_revision_count = content['revision_count']

        response = client.post('/portfolios/13/revisions', {
            'assets': [
                {'asset': 1, 'weight': 40},
                {'asset': 3, 'weight': 60}
            ]
        }, format='json')

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], 1)
        pk = content['last_revision']['pk']

        response = client.put(f'/portfolios/13/revisions/{pk}', {
            'assets': [
                {'asset': 3, 'weight': 10},
                {'asset': 1, 'weight': 90}
            ]
        }, format='json')

        self.assertTrue(200 <= response.status_code < 300, response.content)

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], 1)

    @freeze_time('2021-02-04')
    def test_player_delete(self):
        """Player may not delete old revisions"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        content = parse_json(client.get('/portfolios/12').content)
        initial_revision_count = content['revision_count']
        pk = content['last_revision']['pk']

        response = client.delete(f'/portfolios/12/revisions/{pk}')

        self.assertEqual(response.status_code, 403)

        content = parse_json(client.get('/portfolios/12').content)
        self.assertEqual(content['revision_count'], initial_revision_count)

    @freeze_time('2021-02-04')
    def test_player_delete_today(self):
        """Player may not delete revisions even if they were created today"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        content = parse_json(client.get('/portfolios/13').content)
        initial_revision_count = content['revision_count']

        response = client.post('/portfolios/13/revisions', {
            'assets': [
                {'asset': 1, 'weight': 40},
                {'asset': 3, 'weight': 60},
            ]
        }, format='json')

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], initial_revision_count + 1)
        pk = content['last_revision']['pk']

        response = client.delete(f'/portfolios/13/revisions/{pk}')
        self.assertEqual(response.status_code, 403)

        content = parse_json(client.get('/portfolios/13').content)
        self.assertEqual(content['revision_count'], initial_revision_count + 1)

    @freeze_time('2021-02-04')
    def test_admin_create_validate_min_weight(self):
        """POST-related test"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/portfolios/9/revisions', {
            'assets': [
                {'asset': 4, 'weight': -101},
                {'asset': 5, 'weight': 100},
                {'asset': 7, 'weight': 100},
                {'asset': 9, 'weight': 1},
            ]
        }, format='json')
        self.assertEqual(response.status_code, 400, "Weights may not exceed -100")

    @freeze_time('2021-02-04')
    def test_admin_create_validate_max_weight(self):
        """POST-related test"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/portfolios/9/revisions', {
            'assets': [
                {'asset': 4, 'weight': 101},
                {'asset': 5, 'weight': -1},
            ]
        }, format='json')
        self.assertEqual(response.status_code, 400, "Weights may not exceed 100")

    @freeze_time('2021-02-04')
    def test_admin_create_validate_zero_weight(self):
        """POST-related test"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/portfolios/9/revisions', {
            'assets': [
                {'asset': 4, 'weight': 100},
                {'asset': 5, 'weight': 0},
            ]
        }, format='json')
        self.assertEqual(response.status_code, 400, "Weights may not be zero")

    @freeze_time('2021-02-04')
    def test_admin_create_validate_summed_weight(self):
        """POST-related test"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/portfolios/9/revisions', {
            'assets': [
                {'asset': 4, 'weight': 10},
                {'asset': 5, 'weight': 20}
            ]
        }, format='json')
        self.assertEqual(response.status_code, 400, "Sum of weights must be 100")

    @freeze_time('2021-02-04')
    def test_admin_create_validate_unique_asset(self):
        """POST-related test"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/portfolios/9/revisions', {
            'assets': [
                {'asset': 4, 'weight': 50},
                {'asset': 4, 'weight': 50}
            ]
        }, format='json')
        self.assertEqual(response.status_code, 400, "Assets must only appear once")

    @freeze_time('2021-02-04')
    def test_admin_create_validate_valid_asset(self):
        """Should reject assets that do not belong to this game"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.post('/portfolios/9/revisions', {
            'assets': [
                {'asset': 1, 'weight': 50},
                {'asset': 8, 'weight': 50}
            ]
        }, format='json')
        self.assertEqual(response.status_code, 400, response.content)

    @freeze_time('2021-02-04')
    def test_admin_create_revision_in_empty_portfolio(self):
        """POST-related test"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        # Test creating a revision on an empty portfolio, by creating a new
        # portfolio first:
        response = client.post('/participations/1/portfolios', {
            'name': "Testportfolio",
        }, format='json')
        self.assertTrue(200 <= response.status_code < 300, response.content)

        content = parse_json(response.content)

        path = f'/portfolios/{content["pk"]}/revisions'
        for _ in range(3):
            response = client.post(path, {
                'assets': [
                    {'asset': 4, 'weight': 40},
                    {'asset': 5, 'weight': 60}
                ]
            }, format='json')
            self.assertTrue(200 <= response.status_code < 300, response.content)

        content = parse_json(
            client.get(f'{path}?ordering=date').content
        )

        dates = [revision['date'] for revision in content]

        # Test if the dates are unique
        self.assertEqual(len(set(dates)), len(content))

        # Test if the dates are sorted
        self.assertEqual(sorted(dates), dates)

    def test_results(self):
        """Fetch results for this portfolio revision"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/portfolio-revisions/38/results')
        self.assertEqual(response.status_code, 200, response.content)
        content = parse_json(response.content)

        self.assertTrue(all(
            result['date'] <= '2021-02-05'
            for result in content
        ))
        self.assertEqual(content, [
            {
                "date": "2021-01-05",
                "performance": 0.08934745518227549
            }, {
                "date": "2021-01-06",
                "performance": -0.005495442118975458,
                "sharpe_ratio": 0.8841148409559821,
                "risk": 0.047421448650625475
            }, {
                "date": "2021-01-07",
                "performance": 0.11283010784280512,
                "sharpe_ratio": 1.2817167402761922,
                "risk": 0.0511506988311429
            }, {
                "date": "2021-01-08",
                "performance": 0.10333144144992845,
                "sharpe_ratio": 1.5883605019278952,
                "risk": 0.04722063442019111
            }, {
                "date": "2021-01-11",
                "performance": -0.014241498767097571,
                "sharpe_ratio": 1.033519371896088,
                "risk": 0.05530076578335642
            }, {
                "date": "2021-01-12",
                "performance": 0.014097389216147782,
                "sharpe_ratio": 0.9434953669895568,
                "risk": 0.05297136995346139
            }, {
                "date": "2021-01-13",
                "performance": -0.07373770348405628,
                "sharpe_ratio": 0.49383116894579676,
                "risk": 0.06541615363455403
            }, {
                "date": "2021-01-14",
                "performance": -0.06036649737660538,
                "sharpe_ratio": 0.302768346427425,
                "risk": 0.06843732753952074
            }, {
                "date": "2021-01-15",
                "performance": -0.020563212925787333,
                "sharpe_ratio": 0.24513562941356914,
                "risk": 0.06581483046571104
            }, {
                "date": "2021-01-18",
                "performance": -0.0693949888100498,
                "sharpe_ratio": 0.11230002905242036,
                "risk": 0.06750403436957186
            }, {
                "date": "2021-01-19",
                "performance": -0.12590534189646085,
                "sharpe_ratio": -0.06077846220713348,
                "risk": 0.07493427750642752
            }, {
                "date": "2021-01-20",
                "performance": 0.06564455983280304,
                "sharpe_ratio": 0.017431377177159615,
                "risk": 0.07432128467211
            }, {
                "date": "2021-01-21",
                "performance": 0.09895340766594307,
                "sharpe_ratio": 0.11589091372734842,
                "risk": 0.07599963695849672
            }, {
                "date": "2021-01-22",
                "performance": 0.10693094294542371,
                "sharpe_ratio": 0.20415608677794297,
                "risk": 0.07747245266074294
            }, {
                "date": "2021-01-25",
                "performance": 0.0011288574774988014,
                "sharpe_ratio": 0.19800193738893043,
                "risk": 0.07493511735918845
            }, {
                "date": "2021-01-26",
                "performance": -0.10681179187298895,
                "sharpe_ratio": 0.0923872205647287,
                "risk": 0.07830336520927984
            }, {
                "date": "2021-01-27",
                "performance": -0.01266627812620591,
                "sharpe_ratio": 0.07966948008309344,
                "risk": 0.07610959836969854
            }, {
                "date": "2021-01-28",
                "performance": -0.00679958519549162,
                "sharpe_ratio": 0.07226031275659535,
                "risk": 0.07402389850907358
            }, {
                "date": "2021-01-29",
                "performance": 0.0350577483023935,
                "sharpe_ratio": 0.09553828896635148,
                "risk": 0.07235433026380797
            }, {
                "date": "2021-02-01",
                "performance": 0.06974963007091661,
                "sharpe_ratio": 0.13995684204358136,
                "risk": 0.07183971732864584
            }, {
                "date": "2021-02-02",
                "performance": 0.07867063332269586,
                "sharpe_ratio": 0.18602103400146125,
                "risk": 0.07161500240239209
            }, {
                "date": "2021-02-03",
                "performance": 0.0016014110336972642,
                "sharpe_ratio": 0.18267329634561286,
                "risk": 0.07001103992693099
            }, {
                "date": "2021-02-04",
                "performance": 0.07369580995434122,
                "sharpe_ratio": 0.22183295249486085,
                "risk": 0.0695895854594101
            }, {
                "date": "2021-02-05",
                "performance": -0.055786835624466846,
                "sharpe_ratio": 0.17917313877097832,
                "risk": 0.06959521112546513
            }
        ])

    def test_results_admin(self):
        """Admin can fetch results for any portfolio revision"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='hvddool'))

        response = client.get('/portfolio-revisions/38/results')
        self.assertEqual(response.status_code, 200, response.content)

        self.assertTrue(any(
            result['date'] > '2021-02-05'
            for result in parse_json(response.content)
        ))

    def test_results_other(self):
        """Player may not see results of someone else's portfolio revision"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000920'))

        response = client.get('/portfolio-revisions/16/results')
        self.assertEqual(response.status_code, 404, response.content)

    def test_results_blocked_user(self):
        """Blocked players should not be able to see the revision results"""

        client = self.client
        client.force_authenticate(user=User.objects.get(username='u000902'))

        response = client.get('/portfolio-revisions/60/results')
        self.assertEqual(response.status_code, 404, response.content)
