#! /usr/bin/python3

import datetime
import json
import random
import uuid
import itertools


class Random(random.Random):
    """
    A variant of Random that is seeded using a list of seeds and has
    some additional useful functionality.

    The idea is to use separate deterministic random number generators
    as much as possible, to prevent small changes (added fields, etcetera)
    from causing huge avalanches of cascading changes further down the line.
    """

    def __init__(self, *seeds):
        super().__init__()
        self.__seeds = seeds
        self.seed(repr(seeds))

    def __getitem__(self, args):
        """
        Create a new PRNG with a seed that is unique to both this PRNG
        and the given arguments.
        """
        return type(self)(*self.__seeds, args)

    def random_subset(self, iterator, *, odds=3):
        """
        A subset of an iterator where items have a random chance to be omitted.
        If odds is 3 or absent, each element has a 1/3 chance to be omitted.
        """
        return filter(lambda _: self.randrange(odds), iterator)


class Object(dict):
    """
    Simulate a Javascript object, that is, an object whose attributes
    are also its dict keys. Do not use except in standalone scripts, as it
    creates objects with reference cycles which are hard to garbage-collect.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


def main():
    with open('investment_game/stocks/fixtures/stocks.json') as fh:
        stocks_fixtures = json.load(fh, object_hook=Object)

    indexes = {
        model.pk: model
        for model in stocks_fixtures
        if model.model == 'stocks.Index'
    }

    test_fixtures = []
    demo_fixtures = []
    fixtures = None

    epoch = datetime.date(2021, 1, 1)
    prng = Random()

    sequences = {}

    def add_fixture(model, fields):
        pk = sequences.get(model, 0) + 1
        sequences[model] = pk
        fixtures.append({'model': model, 'pk': pk, 'fields': fields})
        return pk

    for game_index in range(10):
        fixtures = test_fixtures if game_index < 3 else demo_fixtures

        game_prng = prng['game', game_index]
        start_date = epoch + datetime.timedelta(
            days=game_prng['start_date'].randrange(270),
        )
        game_days = game_prng['end_date'].randrange(7, 90)
        game_duration = datetime.timedelta(days=game_days)
        end_date = start_date + game_duration
        embargo_prng = game_prng['embargo']
        if embargo_prng.randrange(3):
            embargo_days = (
                game_prng['embargo_days']
                .randrange(game_days // 2, game_days - 1)
            )
            embargo_date = start_date + datetime.timedelta(days=embargo_days)
        else:
            embargo_days = None
            embargo_date = None
        if game_index:
            game_index_pk = game_index % len(indexes) + 1
            game_asset_pks = indexes[game_index_pk].fields.assets
        else:
            game_index_pk = None
            game_asset_pks = list(
                game_prng['game_asset_pks'].random_subset(range(1, 11))
            )
        max_portfolios = game_prng['max_portfolios'].randrange(1, 10)
        max_portfolio_revisions = (
            game_prng['max_portfolio_revisions']
            .randrange(1, min(game_days - 1, 10))
        )
        max_portfolio_assets = (
            game_prng['max_portfolio_assets']
            .randrange(1, len(game_asset_pks) + 1)
        )
        game_pk = add_fixture('game.Game', {
            'name': f"Game #{game_index + 1}",
            'token': str(uuid.uuid5(
                uuid.NAMESPACE_URL,
                f"https://investment-game.uvt.nl/game/{game_index}",
            )),
            'start_date': str(start_date),
            'embargo_date': str(embargo_date) if embargo_date else None,
            'end_date': str(end_date),
            'index': game_index_pk,
            'assets': game_asset_pks,
            'max_portfolios': max_portfolios,
            'max_portfolio_revisions': max_portfolio_revisions,
            'max_portfolio_assets': max_portfolio_assets,
        })

        if game_index < 3:
            players = game_prng['user_pks'].random_subset(range(5, 10), odds=2)
        else:
            players = game_prng['user_pks'].random_subset(range(1, 35), odds=5)

        for participation_index, user_pk in enumerate(players):
            participation_prng = game_prng[
                'participation', participation_index,
            ]

            participation_pk = add_fixture('game.Participation', {
                'user': user_pk,
                'game': game_pk,
                'blocked': user_pk >= 7 and participation_prng['blocked'].randrange(3) == 0,
            })

            for portfolio_index in range(
                1,
                participation_prng['portfolio_indexes'].randrange(1, max_portfolios + 1) + 1
            ):
                portfolio_pk = add_fixture('game.Portfolio', {
                    'participation': participation_pk,
                    'name':
                        f"Portfolio #{portfolio_index}",
                })

                portfolio_prng = participation_prng[
                    'portfolio', portfolio_index,
                ]

                portfolio_revision_dates = sorted(
                    start_date + datetime.timedelta(days=offset)
                    for offset in portfolio_prng['dates'].sample(
                        range(1, embargo_days or game_days),
                        portfolio_prng['dates'].randrange(max_portfolio_revisions),
                    )
                )
                for portfolio_revision_index, portfolio_revision_date in (
                    enumerate(itertools.chain([start_date], portfolio_revision_dates))
                ):
                    portfolio_revision_prng = portfolio_prng[
                        'portfolio_revision', portfolio_revision_index,
                    ]

                    revision_created_prng = portfolio_revision_prng['created_at']
                    portfolio_revision_time = datetime.datetime(
                        year=portfolio_revision_date.year,
                        month=portfolio_revision_date.month,
                        day=portfolio_revision_date.day,
                        hour=revision_created_prng['hour'].randrange(6, 24),
                        minute=revision_created_prng['minute'].randrange(60),
                        second=revision_created_prng['second'].randrange(60),
                    ).astimezone()

                    portfolio_revision_pk = add_fixture(
                        'game.PortfolioRevision',
                        {
                            'portfolio': portfolio_pk,
                            'created_at': portfolio_revision_time.isoformat(),
                            'date': str(portfolio_revision_date),
                        },
                    )

                    asset_pks = list(game_asset_pks)
                    portfolio_revision_prng['asset_pks'].shuffle(asset_pks)
                    asset_pks = asset_pks[
                        :
                        portfolio_revision_prng['asset_pks_num']
                        .randrange(1, max_portfolio_assets + 1)
                    ]

                    weights = []
                    previous_midpoint = 0
                    for midpoint in portfolio_revision_prng[
                        'weights',
                    ].sample(range(1, 100), len(asset_pks) - 1):
                        weights.append(midpoint - previous_midpoint)
                        previous_midpoint = midpoint
                    weights.append(100 - previous_midpoint)
                    assert sum(weights) == 100
                    assert len(weights) == len(asset_pks)

                    for asset_pk, weight in zip(asset_pks, weights):
                        add_fixture('game.PortfolioAsset', {
                            'portfolio_revision': portfolio_revision_pk,
                            'asset': asset_pk,
                            'weight': weight,
                        })

    test_fixtures.sort(key=lambda elem: (elem['model'], elem['pk']))

    with open('investment_game/game/fixtures/game.json', 'w') as file_handle:
        json.dump(test_fixtures, file_handle, indent=2)

    demo_fixtures.sort(key=lambda elem: (elem['model'], elem['pk']))

    with open('investment_game/game/fixtures/game-extra.json', 'w') as file_handle:
        json.dump(demo_fixtures, file_handle, indent=2)


if __name__ == '__main__':
    main()
