#! /usr/bin/python3

import datetime
import json
import random


class Random(random.Random):
    """
    A variant of Random that is seeded using a list of seeds and has
    some additional useful functionality.

    The idea is to use separate deterministic random number generators
    as much as possible, to prevent small changes (added fields, etcetera)
    from causing huge avalanches of cascading changes further down the line.
    """

    def __init__(self, *seeds):
        super().__init__()
        self.__seeds = seeds
        self.seed(repr(seeds))

    def __getitem__(self, args):
        """
        Create a new PRNG with a seed that is unique to both this PRNG
        and the given arguments.
        """
        return type(self)(*self.__seeds, args)

    def random_subset(self, iterator, *, odds=3):
        """
        A subset of an iterator where items have a random chance to be omitted.
        If odds is 3 or absent, each element has a 1/3 chance to be omitted.
        """
        return filter(lambda _: self.randrange(odds), iterator)


def main():
    test_fixtures = []
    demo_fixtures = []
    fixtures = None

    epoch = datetime.date(2021, 1, 1)
    prng = Random()

    sequences = {}

    def add_fixture(model, fields):
        pk = sequences.get(model, 0) + 1
        sequences[model] = pk
        fixtures.append({'model': model, 'pk': pk, 'fields': fields})
        return pk

    for asset_pk in range(1, 11):
        asset_prng = prng['asset', asset_pk]
        current_price = 100 ** (asset_prng.random() * 1.5)
        asset_drift = asset_prng.gauss(0.0, 0.0002)
        asset_volatility = abs(asset_prng.gauss(0.0, 0.05))

        for day in range(365):
            fixtures = test_fixtures if day < 45 else demo_fixtures

            date = epoch + datetime.timedelta(days=day)
            time = datetime.datetime(
                year=date.year,
                month=date.month,
                day=date.day,
                hour=17,
                minute=0,
                second=0,
            ).astimezone()

            if date.weekday() > 4:
                add_fixture('stocks.AssetPrice', {
                    'asset': asset_pk,
                    'currency': 'EUR',
                    'closing_price': None,
                    'closing_time': time.isoformat(),
                    'closing_date': date.isoformat(),
                    'timezone_name': 'Europe/Amsterdam',
                })
            else:
                current_price *= asset_prng.lognormvariate(
                    asset_drift,
                    asset_volatility,
                )

                add_fixture('stocks.AssetPrice', {
                    'asset': asset_pk,
                    'currency': 'EUR',
                    'closing_price': current_price,
                    'closing_time': time.isoformat(),
                    'closing_date': date.isoformat(),
                    'timezone_name': 'Europe/Amsterdam',
                })

    test_fixtures.sort(key=lambda elem: (elem['model'], elem['pk']))

    with open('investment_game/stocks/fixtures/prices.json', 'w') as file_handle:
        json.dump(test_fixtures, file_handle, indent=2)

    demo_fixtures.sort(key=lambda elem: (elem['model'], elem['pk']))

    with open('investment_game/stocks/fixtures/prices-extra.json', 'w') as file_handle:
        json.dump(demo_fixtures, file_handle, indent=2)


if __name__ == '__main__':
    main()
